@@ -35,7 +35,7 @@ from .abstract_builder import AbstractBuilder
 from .abstract_project import AbstractProject
 from .bindings import Bindings
 from .configurable import Configurable, Option
-from .exceptions import UserException
+from .exceptions import deprecated, UserException
 from .module import resolve_abi_version
 from .py_versions import OLDEST_SUPPORTED_MINOR
 from .pyproject import PyProjectException, PyProjectOptionException
@@ -577,10 +577,11 @@ class Project(AbstractProject, Configurable):
         # parts of the configuration.
         self._configure_from_arguments(tool, tool_description)
 
-        # Now that any help has been given we can report a problematic
+        # Now that any help has been given we can report a missing
         # pyproject.toml file.
-        if pyproject.toml_error:
-            raise PyProjectException(pyproject.toml_error)
+        if pyproject.pyproject is None:
+            raise PyProjectException(
+                    "there is no such file in the current directory")
 
         # Make sure the configuration is complete.
         self.apply_user_defaults(tool)
@@ -667,6 +668,16 @@ class Project(AbstractProject, Configurable):
         # minor version.)
         self.abi_version = resolve_abi_version(self.abi_version, module=False)
 
+        # These ABI versions are deprecated because we have deprecated any
+        # arguments to 'throw()' which these versions rely on.
+        abi_version = tuple([int(v) for v in self.abi_version.split('.')])
+
+        if abi_version < (12, 9):
+            self._deprecated_abi_version('12.9')
+
+        if abi_version == (13, 0):
+            self._deprecated_abi_version('13.1')
+
         # Checks for standalone projects.
         if tool in Option.BUILD_TOOLS and not self.sip_module:
             # Check there is only one set of bindings.
@@ -717,7 +728,8 @@ class Project(AbstractProject, Configurable):
         from argparse import SUPPRESS
         from .argument_parser import ArgumentParser
 
-        parser = ArgumentParser(tool_description, argument_default=SUPPRESS)
+        parser = ArgumentParser(tool_description, build_tool=True,
+                argument_default=SUPPRESS)
 
         # Add the user configurable options to the parser.
         all_options = {}
@@ -758,6 +770,11 @@ class Project(AbstractProject, Configurable):
 
         return int(parts[0]), int(parts[1])
 
+    def _deprecated_abi_version(self, instead):
+        """ Issue a deprecation warning about an old ABI version. """
+
+        deprecated(f"ABI v{self.abi_version}", instead=f"v{instead} or later")
+
     def _enable_disable_bindings(self):
         """ Check the enabled bindings are valid and remove any disabled ones.
         """
